//
// FILE: SimGUI.cpp
//
// Copyright (c) 1997 by Aaron Michael Cohen and Mike Woodring
//
/////////////////////////////////////////////////////////////////////////
#include "SimGUI.h"
#include "resource.h"
#include <math.h>

// declare the single global application instance...
CDiningSim g_theApp;

// CDiningSim member functions...
BOOL CDiningSim::InitInstance(void) {
    m_pMainWnd = new CDiningSimWindow();
    m_pMainWnd->ShowWindow(m_nCmdShow);
    m_pMainWnd->UpdateWindow();
    return TRUE;
}

// CDiningSimWindow MFC message map...
BEGIN_MESSAGE_MAP(CDiningSimWindow, CFrameWnd)
    ON_WM_PAINT()
    ON_WM_ERASEBKGND()
    ON_WM_SIZE()
    ON_WM_VSCROLL()
    ON_WM_HSCROLL()
    ON_WM_GETMINMAXINFO()
    ON_WM_DESTROY()
    ON_COMMAND( ID_FILE_RUN, OnRun)
    ON_COMMAND( ID_FILE_STOP, OnStop)
    ON_COMMAND( ID_FILE_EXIT, OnExit)
    ON_UPDATE_COMMAND_UI( ID_FILE_RUN, OnUpdateRunUI)
    ON_UPDATE_COMMAND_UI( ID_FILE_STOP, OnUpdateStopUI)
    ON_MESSAGE( WM_DININGSIM_PHILOSOPHER_UPDATE, OnDiningSimPhilosopherUpdate)
    ON_MESSAGE( WM_DININGSIM_FORK_UPDATE, OnDiningSimForkUpdate)
END_MESSAGE_MAP()

// CDiningSimWindow constuctor...
CDiningSimWindow::CDiningSimWindow() {
    // not initially running...
    m_bRunning = FALSE;

    // no previous user message text rect...
    m_crLastTextRect = CRect(0,0,0,0);

    // no initial scroll...
    m_nHScrollPos = 0;
    m_nVScrollPos = 0;
    m_nHScrollPosMax = 0;
    m_nVScrollPosMax = 0;

    // default max window size is a big as the screen...
    CRect dr;
    GetDesktopWindow()->GetClientRect(dr);
    m_ptMaxWindowSize.x = dr.Width();
    m_ptMaxWindowSize.y = dr.Height();

    // table side will be 3/5 of the window width...
    m_nTableSide = 3*WINDOW_WIDTH/5;

    // initialize the philosopher and fork states...
    for (int i = 0; i < NUMBER_PHILOSOPHERS; i++) {
        m_nForkState[i] = FORK_STATE_AVAILABLE;
        m_nPhilosopherState[i] = PHILOSOPHER_STATE_THINKING;
    }

    // load the image of the spaghetti bowl...
    HBITMAP hBitmap;
    hBitmap = (HBITMAP) ::LoadImage( AfxGetApp()->m_hInstance, "bowl.bmp", IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);
    if (hBitmap)
        m_cbmBowl.Attach(hBitmap);
    else
        ErrorExit("Unable to load image file 'bowl.bmp'!");

    // load the images of the philosophers...
    hBitmap = (HBITMAP) ::LoadImage( AfxGetApp()->m_hInstance, "philosophers.bmp", IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);
    if (hBitmap)
        m_cbmPhilosophers.Attach(hBitmap);
    else 
        ErrorExit("Unable to load image file 'philosophers.bmp'!");

    // load the images of the forks...
    hBitmap = (HBITMAP) ::LoadImage( AfxGetApp()->m_hInstance, "forks.bmp", IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);
    if (hBitmap)
        m_cbmForks.Attach(hBitmap);
    else 
        ErrorExit("Unable to load image file 'forks.bmp'!");

    // create the main window...
    Create( NULL, "Dining Philosophers Simulation", WS_OVERLAPPEDWINDOW | WS_VSCROLL | WS_HSCROLL, rectDefault, NULL, MAKEINTRESOURCE(IDR_FILEMENU));

    // compute maximum size, resize, and center the main window...
    CRect wr, cr;
    GetClientRect(cr);
    GetWindowRect(wr);
    m_ptMaxWindowSize.x = WINDOW_WIDTH + wr.Width() - cr.Width();
    m_ptMaxWindowSize.y = WINDOW_HEIGHT + wr.Height() - cr.Height();
    int x = (m_ptMaxWindowSize.x > dr.Width()) ? dr.Width() : m_ptMaxWindowSize.x;
    int y = (m_ptMaxWindowSize.y > dr.Height()) ? dr.Height() : m_ptMaxWindowSize.y;
    SetWindowPos( NULL, 0, 0, x, y, SWP_NOMOVE | SWP_NOZORDER);
    CenterWindow();

    // set the simualation UI window...
    m_cSimulation.SetHwnd(m_hWnd);

    // show that we're ready...
    ShowUserMessage("Ready.");
}

// CDiningSimWindow standard message handler functions...
afx_msg void CDiningSimWindow::OnPaint() {
    // redraw the entire simulation...
    CPaintDC dc(this);

    // align the DC origin with the scroll position...
    dc.SetWindowOrg( m_nHScrollPos, m_nVScrollPos);

    // update the display...
    DrawSimulation(&dc);
}

afx_msg BOOL CDiningSimWindow::OnEraseBkgnd(CDC *pDC) {
    // fill the window with black...
    CRect cr;
    CBrush cb;
    GetClientRect(cr);
    cb.CreateStockObject(BLACK_BRUSH);
    pDC->FillRect( cr, &cb);
    return TRUE;
}

afx_msg void CDiningSimWindow::OnSize( UINT nType, int cx, int cy) {
    // adjust the horizontal scroll bar for the new window size...
    if (cx >= WINDOW_WIDTH) {
        // hide the scrollbar...
        SetScrollRange( SB_HORZ, 0, 0, TRUE);
        m_nHScrollPos = 0;
        m_nHScrollPosMax = 0;
    }
    else {
        // set the new scroll range...
        m_nHScrollPosMax = WINDOW_WIDTH - cx;
        SetScrollRange( SB_HORZ, 0, m_nHScrollPosMax, TRUE);

        // make the the current setting is not beyond the 
        // new upper bound...
        if (m_nHScrollPos > m_nHScrollPosMax) { 
            m_nHScrollPos = m_nHScrollPosMax;
            SetScrollPos( SB_HORZ, m_nHScrollPos, TRUE);
        }
    }

    // adjust the vertical scroll bar for the new window size...
    if (cy >= WINDOW_HEIGHT) {
        // hide the scrollbar...
        SetScrollRange( SB_VERT, 0, 0, TRUE);
        m_nVScrollPos = 0;
        m_nVScrollPosMax = 0;
    }
    else {
        // set the new scroll range...
        m_nVScrollPosMax = WINDOW_HEIGHT - cy;
        SetScrollRange( SB_VERT, 0, m_nVScrollPosMax, TRUE);

        // make the the current setting is not beyond the 
        // new upper bound...
        if (m_nVScrollPos > m_nVScrollPosMax) { 
            m_nVScrollPos = m_nVScrollPosMax;
            SetScrollPos( SB_VERT, m_nVScrollPos, TRUE);
        }
    }
}

afx_msg void CDiningSimWindow::OnVScroll( UINT nCode, UINT nPos, CScrollBar *pScrollBar) {
    // use this to store how far to move the window...
    int nDelta = 0;

    // simple scrolling...
    // a page equals a line equals one row of pixels...
    // dragging the thumb allows faster scrolling...
    switch (nCode) {
        case SB_PAGEUP:
        case SB_LINEUP:
            if (m_nVScrollPos <= 0)
                return;
            else
                nDelta = -1;
            break;
        case SB_PAGEDOWN:
        case SB_LINEDOWN:
            if (m_nVScrollPos >= m_nVScrollPosMax)
                return;
            else
                nDelta = 1;
            break;
        case SB_THUMBPOSITION:
            nDelta = (int) nPos - m_nVScrollPos;
            break;
    }

    // save the new scroll position...
    m_nVScrollPos += nDelta;

    // set the scroll bar to reflect the new position...
    SetScrollPos( SB_VERT, m_nVScrollPos, TRUE);

    // scroll the window and redraw the uncovered area...
    ScrollWindow( 0, -nDelta);
}

afx_msg void CDiningSimWindow::OnHScroll( UINT nCode, UINT nPos, CScrollBar *pScrollBar) {
    // use this to store how far to move the window...
    int nDelta = 0;

    // simple scrolling...
    // a page equals a line equals one column of pixels...
    // dragging the thumb allows faster scrolling...
    switch (nCode) {
        case SB_PAGEUP:
        case SB_LINEUP:
            if (m_nHScrollPos <= 0)
                return;
            else
                nDelta = -1;
            break;
        case SB_PAGEDOWN:
        case SB_LINEDOWN:
            if (m_nHScrollPos >= m_nHScrollPosMax)
                return;
            else
                nDelta = 1;
            break;
        case SB_THUMBPOSITION:
            nDelta = (int) nPos - m_nHScrollPos;
            break;
    }

    // save the new scroll position...
    m_nHScrollPos += nDelta;

    // set the scroll bar to reflect the new position...
    SetScrollPos( SB_HORZ, m_nHScrollPos, TRUE);

    // scroll the window and redraw the uncovered area...
    ScrollWindow( -nDelta, 0);
}

afx_msg void CDiningSimWindow::OnGetMinMaxInfo( MINMAXINFO FAR* lpMMI) {
    // tell Windows the application maximum window size...
    lpMMI->ptMaxSize = lpMMI->ptMaxTrackSize = m_ptMaxWindowSize;
}

afx_msg void CDiningSimWindow::OnRun() {
    // start the simulation running...
    ShowUserMessage("Starting...");
    StartSimulation();
    ShowUserMessage("Running.");
}

afx_msg void CDiningSimWindow::OnStop() {
    // tell the simulation to stop...
    ShowUserMessage("Stopping...");
    StopSimulation();
    ShowUserMessage("Ready.");
}

afx_msg void CDiningSimWindow::OnUpdateRunUI( CCmdUI *pCmdUI) {
    // set the state of the RUN menu option...
    pCmdUI->Enable(!m_bRunning);
}

afx_msg void CDiningSimWindow::OnUpdateStopUI( CCmdUI *pCmdUI) {
    // set the state of the STOP menu option...
    pCmdUI->Enable(m_bRunning);
}

afx_msg void CDiningSimWindow::OnExit() {
    // tell the application to terminate...
    PostMessage( WM_CLOSE);
}

afx_msg void CDiningSimWindow::OnDestroy() {
    // ensure the the simulation has stopped before
    // we destroy the window...
    ShowUserMessage("Shutting down...");
    StopSimulation();
    ShowUserMessage("Finished.");
}

// CDiningSimWindow user defined message handler functions...
afx_msg LONG CDiningSimWindow::OnDiningSimPhilosopherUpdate( UINT uPhilosopher, LONG lNewState) {
    // change the philosopher state...
    m_nPhilosopherState[uPhilosopher] = lNewState;

    // get the window device context...
    CDC *pDC = GetDC();

    // align the DC origin with the scroll position...
    pDC->SetWindowOrg( m_nHScrollPos, m_nVScrollPos);
    
    // update the display...    
    DrawPhilosopher( pDC, uPhilosopher);
    ReleaseDC(pDC);

    return 0;
}

afx_msg LONG CDiningSimWindow::OnDiningSimForkUpdate( UINT uFork, LONG lNewState) {
    // change the fork state...
    m_nForkState[uFork] = lNewState;

    // get the window device context...
    CDC *pDC = GetDC();

    // align the DC origin with the scroll position...
    pDC->SetWindowOrg( m_nHScrollPos, m_nVScrollPos);

    // update the display...
    DrawFork( pDC, uFork);
    ReleaseDC(pDC);

    return 0;
}

// CDiningSimWindow simulation control functions...
void CDiningSimWindow::StartSimulation(void) {
    // run the simulation if it is stopped...
    if (!m_bRunning) {
        m_cSimulation.Run();
        m_bRunning = TRUE;
    }
}

void CDiningSimWindow::StopSimulation(void) {
    // stop the simulation if it is running...
    if (m_bRunning) {
        m_cSimulation.Stop();
        m_bRunning = FALSE;
    }
}

void CDiningSimWindow::ShowUserMessage( CString csMessage) {
    // set the user message and update the display...
    m_csUserMessage = csMessage;

    // get the window device context...
    CDC *pDC = GetDC();

    // align the DC origin with the scroll position...
    pDC->SetWindowOrg( m_nHScrollPos, m_nVScrollPos);

    // update the display...
    DrawUserMessage(pDC);
    ReleaseDC(pDC);
}

// CDiningSimWindow drawing functions...
void CDiningSimWindow::DrawSimulation(CDC *pDC) {
    // get the window client area...
    CRect cr;
    GetClientRect(cr);

    // compute the center of the scrollable window...
    CPoint cpt( WINDOW_WIDTH >> 1, WINDOW_HEIGHT >> 1);

    // draw table in center of window...
    CBrush cb;
    cb.CreateStockObject(NULL_BRUSH);
    CBrush *pOldBrush = pDC->SelectObject(&cb);
    CPen cp;
    cp.CreateStockObject(WHITE_PEN);
    CPen *pOldPen = pDC->SelectObject(&cp);
    pDC->Ellipse( cpt.x - m_nTableSide/2, cpt.y - m_nTableSide/2, cpt.x + m_nTableSide/2, cpt.y + m_nTableSide/2);
    pDC->SelectObject(pOldBrush);
    pDC->SelectObject(pOldPen);

    // create a DC to use for drawing bitmaps...
    CDC mdc;
    mdc.CreateCompatibleDC(pDC);

    // used for computing drawing locations...
    POINT pt;

    // used to iterate over the seating positions...
    int nSeat;
    int iRadius;

    // draw the bowls...
    CBitmap *pOldBitmap = mdc.SelectObject(&m_cbmBowl);
    iRadius = m_nTableSide/2 - BOWL_SPACING;
    for (nSeat = 0; nSeat < NUMBER_PHILOSOPHERS; nSeat++) {
        pt = ComputeSeatPosition( cpt, NUMBER_PHILOSOPHERS, nSeat, iRadius);
        pt.x -= BITMAP_WIDTH >> 1;
        pt.y -= BITMAP_HEIGHT >> 1;        
        pDC->BitBlt( pt.x, pt.y, BITMAP_WIDTH, BITMAP_HEIGHT, &mdc, 0, 0, SRCCOPY);
    }
    mdc.SelectObject(pOldBitmap);

    // draw the philosophers...
    for (nSeat = 0; nSeat < NUMBER_PHILOSOPHERS; nSeat++) {
        DrawPhilosopher( pDC, nSeat);
    }

    // draw the forks...
    for (nSeat = 0; nSeat < NUMBER_PHILOSOPHERS; nSeat++) {
        DrawFork( pDC, nSeat);
    }

    // draw the current user message...
    DrawUserMessage( pDC);
}

void CDiningSimWindow::DrawPhilosopher(CDC *pDC, UINT uPhilosopher) {
    // get the window client area...
    CRect cr;
    GetClientRect(cr);

    // compute the center of the scrollable window...
    CPoint cpt( WINDOW_WIDTH >> 1, WINDOW_HEIGHT >> 1);

    // create a DC to use for drawing bitmaps...
    CDC mdc;
    mdc.CreateCompatibleDC(pDC);

    // compute philosopher distance from table center...
    int iRadius = m_nTableSide/2 + PHILOSOPHER_SPACING;

    // compute the upper-left corner of the blit rectangle...
    POINT pt = ComputeSeatPosition( cpt, NUMBER_PHILOSOPHERS, uPhilosopher, iRadius);
    pt.x -= BITMAP_WIDTH >> 1;
    pt.y -= BITMAP_HEIGHT >> 1;        

    // blit the current philosopher image...
    CBitmap *pOldBitmap = mdc.SelectObject(&m_cbmPhilosophers);
    pDC->BitBlt( pt.x, pt.y, BITMAP_WIDTH, BITMAP_HEIGHT, &mdc, BITMAP_WIDTH*m_nPhilosopherState[uPhilosopher], 0, SRCCOPY);
    mdc.SelectObject(pOldBitmap);
}

void CDiningSimWindow::DrawFork(CDC *pDC, UINT uFork) {
    // get the window client area...
    CRect cr;
    GetClientRect(cr);

    // compute the center of the scrollable window...
    CPoint cpt( WINDOW_WIDTH >> 1, WINDOW_HEIGHT >> 1);

    // create a DC to use for drawing bitmaps...
    CDC mdc;
    mdc.CreateCompatibleDC(pDC);

    // compute philosopher distance from table center...
    int iRadius = m_nTableSide/2 + PHILOSOPHER_SPACING;

    // compute the upper-left corner of the blit rectangle...
    POINT pt = ComputeForkPosition( cpt, NUMBER_PHILOSOPHERS, uFork, iRadius);
    pt.x -= BITMAP_WIDTH >> 1;
    pt.y -= BITMAP_HEIGHT >> 1;        

    // blit the current fork image...
    CBitmap *pOldBitmap = mdc.SelectObject(&m_cbmForks);
    pDC->BitBlt( pt.x, pt.y, BITMAP_WIDTH, BITMAP_HEIGHT, &mdc, BITMAP_WIDTH*m_nForkState[uFork], 0, SRCCOPY);
    mdc.SelectObject(pOldBitmap);
}

void CDiningSimWindow::DrawUserMessage(CDC *pDC) {
    // draw the current user message text string m_csUserMessage
    // in the upper-left corner of the window...
    COLORREF crOldTextColor = pDC->SetTextColor(RGB(0,255,0));
    COLORREF crOldBkColor = pDC->SetBkColor(RGB(0,0,0));

    // draw OPAQUE with the previous text rect to make sure that
    // we erase the previous message...
    pDC->ExtTextOut( 0, 0, ETO_OPAQUE, m_crLastTextRect, m_csUserMessage, NULL);

    pDC->SetTextColor(crOldTextColor);
    pDC->SetBkColor(crOldBkColor);

    // save the text rect for later erasing...
    CSize cz = pDC->GetTextExtent(m_csUserMessage);
    m_crLastTextRect = CRect( 0, 0, cz.cx, cz.cy);        
}

// CDiningSimWindow helper functions...
POINT CDiningSimWindow::ComputeSeatPosition( POINT ptCenter, int nPositions, int iPosition, int iRadius) {
    POINT pt;        
    double rad = iPosition * (2 * 3.14159) / nPositions;
    pt.x = (int)(ptCenter.x + sin(rad) * iRadius);
    pt.y = (int)(ptCenter.y - cos(rad) * iRadius);
    return pt;
}

POINT CDiningSimWindow::ComputeForkPosition( POINT ptCenter, int nPositions, int iPosition, int iRadius) {
    POINT pt;        
    double rad = (iPosition - 0.50) * (2 * 3.14159) / nPositions;
    pt.x = (int)(ptCenter.x + sin(rad) * iRadius);
    pt.y = (int)(ptCenter.y - cos(rad) * iRadius);
    return pt;
}

void CDiningSimWindow::ErrorExit( LPCTSTR lpMsg) {
    ::MessageBox( NULL, lpMsg, "DiningSim Fatal Error - Program will Exit", MB_ICONSTOP);
    ::ExitProcess(0xFFFFFFFF);
}
